/* BeaSTM Add-On (c) Thorbjörn Jemander */
#include "PSignal.h"

extern bool SinCosTableExists[];

extern float 	*SIN_TABLE[];
extern float 	*COS_TABLE[];

#define min(x,y) ((x)<(y)?(x):(y))

#pragma export on
extern "C" 
{
bool Process( uint16 *argSourceData, uint16 *argDestinationData, uint32 XSize, uint32 YSize );
}
#pragma export reset

bool	MakeFFTOfLineAndCopyTo( uint32 argLineNumber, float *argDestData, uint16 *argSourceData, uint32 argXSize, uint32 argYSize );
bool	MakeFFTOfColumnAndCopyTo( uint32 argColumnNumber, uint16 *argDestData, float *argSourceData, uint32 argXSize, uint32 argYSize   );

bool
Process( uint16 *argSourceData, uint16 *argDestinationData, uint32 argXSize, uint32 argYSize )
{
	uint32	column,
			line;
	bool	NoErrorSoFar = true;
	
	float	*MiddleBuffer =(float*)calloc( sizeof(float),2*argXSize*argYSize);
	
	for( line = 0; line < argYSize && NoErrorSoFar; line++ )
		NoErrorSoFar = MakeFFTOfLineAndCopyTo(	line,
												MiddleBuffer,
												argSourceData,
												argXSize,
												argYSize );


	for( column = 0; column < argXSize && NoErrorSoFar; column++ )
		NoErrorSoFar = MakeFFTOfColumnAndCopyTo(	column,
													argDestinationData,
													MiddleBuffer,
													argXSize,
													argYSize );
	free( MiddleBuffer );

	return NoErrorSoFar;
}

bool
MakeFFTOfLineAndCopyTo( uint32 argLineNumber, float *argDestData, uint16 *argSourceData, uint32 argXSize, uint32 argYSize  )
{
	int32 	N = argXSize, 
			M,
			i;
			
	static float *R,*I;

	if( !R )
		R = (float *)calloc( sizeof(float),N);
	if( !I )
		I = (float *)calloc( sizeof(float),N);
	
	if( ! I || ! R )
		return false;

	memset( I,0,sizeof(float)*N);

	for( i=0; i<N; i++ )
		R[i] = (float)argSourceData[ argLineNumber * argXSize + i];

	M = CalcStageDepth( N );				// N = 2^M
	
	if( SinCosTableExists[ M ] == false )			// Make twiddle factors if
		if( MakeSineAndCosTable( M ) == false )		// they don't already exist.
			return false;
	
	CooleyTukeyFFT( R,I,N);

//	for( i=0; i<N; i++ )
//		argDestData[ argLineNumber * argXSize + i ] = 8*(uint16)(min(sqrt((R[i]*R[i] + I[i]*I[i])/(4*argXSize*argXSize)),65535) );
//
//	if( argLineNumber == 15 )
//		for( i=0; i<N; i++ )
//			printf("%d argDestData %d\n", i,argDestData[ argLineNumber * argXSize + i ]);		


	for( i=0; i<N/2; i++ )
	{
		argDestData[ argLineNumber * argXSize + i ] = R[i+N/2];
		argDestData[ argLineNumber * argXSize + i + argXSize*argYSize] = I[i+N/2];
	}
	
	for( i=N/2; i<N; i++ )
	{
		argDestData[ argLineNumber * argXSize + i ] = R[i-N/2];
		argDestData[ argLineNumber * argXSize + i + argXSize*argYSize] = I[i-N/2];
	}

	return true;
}

bool
MakeFFTOfColumnAndCopyTo( uint32 argColumnNumber, uint16 *argDestData, float *argSourceData, uint32 argXSize, uint32 argYSize   )
{
	int32 	N = argXSize, 
			M,
			i;
			
	static float *R,*I;

	if( !R )
		R = (float *)calloc( sizeof(float),N);
	if( !I )
		I = (float *)calloc( sizeof(float),N);
	
	if( ! I || ! R )
		return false;

	for( i=0; i<N; i++ )
	{
		R[ i ] = argSourceData[ argColumnNumber + i*argXSize ];
		I[ i ] = argSourceData[ argColumnNumber + i*argXSize + argXSize * argYSize ];
	}
	
	M = CalcStageDepth( N );				// N = 2^M
	
	if( SinCosTableExists[ M ] == false )			// Make twiddle factors if
		if( MakeSineAndCosTable( M ) == false )		// they don't already exist.
			return false;

	CooleyTukeyFFT( R,I,N);

	for( i=0; i<N/2; i++ )
		argDestData[ argColumnNumber + argXSize * i ] = 8*(uint16)(min(sqrt((R[i+N/2]*R[i+N/2] + I[i+N/2]*I[i+N/2])/(4*argXSize*argXSize)),65535) );

	for( i=N/2; i<N; i++ )
		argDestData[ argColumnNumber + argXSize * i ] = 8*(uint16)(min(sqrt((R[i-N/2]*R[i-N/2] + I[i-N/2]*I[i-N/2])/(4*argXSize*argXSize)),65535) );

	return true;
}
