/***************************************************************************\
*																			*
*	Filename: 	Signal.cpp													*
*	Contains: 	Contains some signal processing routines.					*
*	Created:	19:35 2 Sep 96												*
*	Modified:	19:35 2 Sep 96												*
*	Authors:	Thorbjrn Andersson											*
*	Version:	1.0															*
*	Known bugs:	None.														*
*																			*
\***************************************************************************/

#ifndef _P_SIGNAL_H
#define _P_SIGNAL_H

#include "PSignal.h"

bool SinCosTableExists[32] = { FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE};
bool DigitReverseTableExists[32] = { FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE,FALSE};

float 	*SIN_TABLE[32];
float 	*COS_TABLE[32];
short	*DIGIT_REVERSE_TABLE[32];

void
CooleyTukeyFFT( float *X_Re, float *X_Im, int N)
{
	int M, M1, Stage, q, p, k, Ns, kNs;
	
	float TempRe, TempIm, WSin, WCos, TwoPiN;

	M = CalcStageDepth( N );

	if( SinCosTableExists[ M ] == FALSE )
		if( MakeSineAndCosTable( M ) == FALSE )
			return;

	Ns 		= N;
	M1		= M;
	TwoPiN	= 2*PI/N;
	
	for( Stage =1; Stage<= M; Stage++)
	{
		k		=	0;
		Ns 		>>= 1;
		M1--;

		while( k< N )
		{
			for( q=1; q<=Ns; q++ )
			{
				p 			= DigitReverse( k >> M1, M );
				WCos 		= COS_TABLE[ M ][ p ];
				WSin 		= -SIN_TABLE[ M ][ p ];
				kNs  		= k+Ns;
				TempRe 		= X_Re[ kNs ]*WCos - X_Im[ kNs ]*WSin;
				TempIm 		= X_Im[ kNs ]*WCos + X_Re[ kNs ]*WSin;
				X_Re[ kNs ] = X_Re[ k ] - TempRe;
				X_Im[ kNs ]	= X_Im[ k ] - TempIm;
				X_Re[ k ] 	= X_Re[ k ] + TempRe;
				X_Im[ k ]	= X_Im[ k ] + TempIm;
				k++;
			}
			k+=Ns;
		}
	}

	Unscramble( X_Re, X_Im, N );
}

int
DigitReverse( int digit, int M )
{
	long q,newdigit=0;

	for( q=1; q<=M; q++)
	{
		newdigit += digit & 1;
		newdigit <<= 1;
		digit	 >>= 1;
	}
	return newdigit>>1;
}


bool
IsRadix2( int len )
{
	int i, NumberOfBits=0;
	
	for( i=0; i<32; i++)
	{
		NumberOfBits += len & 1;
		len			>>= 1;
	}	
	if( NumberOfBits == 1)
		return TRUE;
	else
		return FALSE;
}

int 
CalcStageDepth( int len )
{
	// Presumes radix-2 length!!! Maximum depth is 32.

	int i;

	for( i=0; i<32; i++)
	{
		if( len & 1)
			return i;
		len >>= 1;
	}
	return 32;
}	

void
Unscramble( float *XR, float *XI, int N)
{
	float TempR, TempI;

	int k,q, M = CalcStageDepth( N );

	for( k=0; k<=N-1; k++)
	{
		q = DigitReverse(k,M);
		if( q>k )
		{
			TempR	= XR[k];
			TempI	= XI[k];
			XR[k]	= XR[q];
			XI[k]	= XI[q];
			XR[q]	= TempR;
			XI[q]	= TempI;
		}
	}
}


bool
MakeSineAndCosTable( int BitSize )
{
	long i,N;

	float TwoPiN;

	if( BitSize >32 || BitSize < 0)
		return FALSE;
	
	N = 1 << BitSize;

	SIN_TABLE[ BitSize ] = (float *)calloc( sizeof(float), N );

	if( SIN_TABLE[ BitSize ] == NULL )
		return FALSE;

	COS_TABLE[ BitSize ] = (float *)calloc( sizeof(float), N );
		
	if( COS_TABLE[ BitSize ] == NULL )
	{
		free( SIN_TABLE[ BitSize ] );
		return FALSE;
	}

	TwoPiN = 2*PI/N;

	for( i=0; i<N; i++)
	{
		SIN_TABLE[ BitSize ][ i ] = sin( TwoPiN*i );
		COS_TABLE[ BitSize ][ i ] = cos( TwoPiN*i );
	}
	
	SinCosTableExists[ BitSize ] = TRUE;

	return TRUE;

}

void
FreeSinAndCosTables()
{
	for( int i=0; i<32; i++)
		if( SinCosTableExists[i] ) 
		{
			free( SIN_TABLE[i] );
			free( COS_TABLE[i] );
		}
}

bool
MakeDigitReverseTable( int BitSize )
{
	long i,N;

	if( BitSize >32 || BitSize < 0)
		return FALSE;
		
	N = 1 << BitSize;

	DIGIT_REVERSE_TABLE[ BitSize ] = (short *)calloc( sizeof(short), N );

	if( SIN_TABLE[ BitSize ] == NULL )
		return FALSE;

	for( i=0; i<N; i++)
		SIN_TABLE[ BitSize ][ i ] = DigitReverse(i,BitSize);
	
	DigitReverseTableExists[ BitSize ] = TRUE;

	return TRUE;

}


#endif